from torch import nn, Tensor
from typing import Union


def mean_with_attn(x: Tensor, mask: Tensor | None, dim: int) -> Tensor:
    """Compute mean along dimension, ignoring masked positions."""
    if mask is None:
        return x.mean(dim=dim)
    # Clone to avoid modifying input
    x = x.clone()
    # Expand mask to match x dimensions if needed
    while mask.ndim < x.ndim:
        mask = mask.unsqueeze(-1)
    mask = mask.expand_as(x)
    x[mask] = 0
    count = (~mask).sum(dim=dim, keepdim=True).float()
    return x.sum(dim=dim) / count.squeeze(dim)


class PatchwisePredictor(nn.Module):
    _MASK_TOKEN = 0
    def __init__(
        self,
        latent_proj,
        proj_over: Union['mean_spec', 'all_spec'] | None = 'mean_spec',
        num_spec_patches: int | None = None,
        apply_mask: bool = False,
    ):
        super().__init__()
        self.latent_proj = latent_proj
        self.proj_over = proj_over
        self.apply_mask = apply_mask

        if proj_over not in ['mean_spec', 'all_spec', None]:
            raise NotImplementedError(f'proj_over = {self.proj_over} not implemented')

    def forward(
        self,
        Z,
        attn_mask=None
    ):
        B, T, M, E = Z.shape
        if self.apply_mask and attn_mask is not None and attn_mask.any():
            Z = Z.reshape(B, -1, E)
            # Require clone for backprop
            Z = Z.clone()
            Z[attn_mask] = self._MASK_TOKEN
            Z = Z.reshape(B, T, M, E)
        # Z : (B, T, M, E)

        if self.proj_over == 'all_spec':
            Z = Z.reshape(B, T, -1)
            # (B, T, M*E)

        # Project from latent embedding space to target space
        logits = self.latent_proj(Z) # E -> C
        # (B, T, M, C)

        if self.proj_over == 'mean_spec':
            if attn_mask is not None:
                attn_mask = attn_mask.reshape(B, T, M)
            logits = mean_with_attn(logits, attn_mask, dim=-2)
            # (B, T, C)

        return logits

    def extra_repr(self):
        return f'proj_over={self.proj_over}'


# --------------------------------------------------------
# BEATs: Audio Pre-Training with Acoustic Tokenizers (https://arxiv.org/abs/2212.09058)
# Github source: https://github.com/microsoft/unilm/tree/master/beats
# Copyright (c) 2022 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------


import torch
import torch.nn as nn
from torch.nn import LayerNorm
import torchaudio.compliance.kaldi as ta_kaldi

from .backbone import (
    TransformerEncoder,
)

import logging
from typing import Optional, Literal

logger = logging.getLogger(__name__)


class BEATsConfig:
    def __init__(self, cfg=None):
        self.input_patch_size: int = -1  # path size of patch embedding
        self.embed_dim: int = 512  # patch embedding dimension
        self.conv_bias: bool = False  # include bias in conv encoder

        self.encoder_layers: int = 12  # num encoder layers in the transformer
        self.encoder_embed_dim: int = 768  # encoder embedding dimension
        self.encoder_ffn_embed_dim: int = 3072  # encoder embedding dimension for FFN
        self.encoder_attention_heads: int = 12  # num encoder attention heads
        self.activation_fn: str = "gelu"  # activation function to use

        self.layer_wise_gradient_decay_ratio: float = 1.0  # ratio for layer-wise gradient decay
        self.layer_norm_first: bool = False  # apply layernorm first in the transformer
        self.deep_norm: bool = False  # apply deep_norm first in the transformer

        # dropouts
        self.dropout: float = 0.1  # dropout probability for the transformer
        self.attention_dropout: float = 0.1  # dropout probability for attention weights
        self.activation_dropout: float = 0.0  # dropout probability after activation in FFN
        self.encoder_layerdrop: float = 0.0  # probability of dropping a tarnsformer layer
        self.dropout_input: float = 0.0  # dropout to apply to the input (after feat extr)

        # positional embeddings
        self.conv_pos: int = 128  # number of filters for convolutional positional embeddings
        self.conv_pos_groups: int = 16  # number of groups for convolutional positional embedding

        # relative position embedding
        self.relative_position_embedding: bool = False  # apply relative position embedding
        self.num_buckets: int = 320  # number of buckets for relative position embedding
        self.max_distance: int = 1280  # maximum distance for relative position embedding
        self.gru_rel_pos: bool = False  # apply gated relative position embedding

        # label predictor
        self.finetuned_model: bool = False  # whether the model is a fine-tuned model.
        self.predictor_dropout: float = 0.1  # dropout probability for the predictor
        self.predictor_class: int = 527  # target class number for the predictor

        if cfg is not None:
            self.update(cfg)

    def update(self, cfg: dict):
        self.__dict__.update(cfg)


class BEATs(nn.Module):
    def __init__(
            self,
            cfg: BEATsConfig,
    ) -> None:
        super().__init__()
        logger.info(f"BEATs Config: {cfg.__dict__}")

        self.cfg = cfg

        self.embed = cfg.embed_dim
        self.post_extract_proj = (
            nn.Linear(self.embed, cfg.encoder_embed_dim)
            if self.embed != cfg.encoder_embed_dim
            else None
        )

        self.input_patch_size = cfg.input_patch_size
        self.patch_embedding = nn.Conv2d(1, self.embed, kernel_size=self.input_patch_size, stride=self.input_patch_size,
                                         bias=cfg.conv_bias)

        self.dropout_input = nn.Dropout(cfg.dropout_input)

        assert not cfg.deep_norm or not cfg.layer_norm_first
        self.encoder = TransformerEncoder(cfg)
        self.layer_norm = LayerNorm(self.embed)

        if cfg.finetuned_model:
            self.predictor_dropout = nn.Dropout(cfg.predictor_dropout)
            self.predictor = nn.Linear(cfg.encoder_embed_dim, cfg.predictor_class)
        else:
            self.predictor = None

    def forward_padding_mask(
            self,
            features: torch.Tensor,
            padding_mask: torch.Tensor,
            threshold: Literal['any', 'all'] | float = 0.5
    ) -> torch.Tensor:
        if padding_mask.shape[:2] == features.shape[:2] :
            return padding_mask
        
        extra = padding_mask.size(1) % features.size(1)
        if extra > 0:
            # drop up to extra (endpoints dont match)
            padding_mask = padding_mask[:, :-extra]
        # Reshape to match feature shape
        # -1: left over act as downsample/stride
        padding_mask = padding_mask.view(
            padding_mask.size(0), features.size(1), -1
        )
        # if all samples in dim is true, then downsample should be true
        if threshold == 'all':
            # consider mask, only if all subtokens were masked
            padding_mask = padding_mask.all(dim=-1)
        elif threshold == 'any':
            # consider mask, if any of the subtokens were masked
            padding_mask = padding_mask.any(dim=-1)
        else:
            # consider mask, if a percentage of subtokens were masked
            padding_mask = padding_mask.mean(dim=-1, dtype=torch.float) > threshold

        return padding_mask

    def preprocess(
        self,
        source: torch.Tensor,
        *,
        fbank_mean: float = 0,
        fbank_std: float = 0.5,
        num_mel_bins = 128,
        sample_frequency = 16_000,
        frame_length = 25, #ms
        frame_shift = 10, #ms
        **kwargs
    ) -> torch.Tensor:
        fbanks = []
        for waveform in source:
            waveform = waveform.unsqueeze(0) * 2 ** 15
            fbank = ta_kaldi.fbank(waveform, num_mel_bins=num_mel_bins, sample_frequency=sample_frequency, frame_length=frame_length, frame_shift=frame_shift)
            fbanks.append(fbank)
        fbank = torch.stack(fbanks, dim=0)
        fbank = (fbank - fbank_mean) / (2 * fbank_std)
        return fbank

    def forward(self,
                fbank: torch.Tensor,
                padding_mask: Optional[torch.Tensor] = None,
                attn_mask: torch.Tensor | None = None,
                return_encoder_layer: Optional[int] = None,
                ):
        # Divide melspec into patches
        features = self.patch_embedding(fbank)
        # B - Batch
        # C - Channel (default 512)
        # T - Time patches
        # M - Mel frequency patches (default 8)
        B, C, T, M = features.size()

        # Combine mel and time patches
        features = features.reshape(B, C, -1)
        # Swap channel and (mel&time) dim
        features = features.transpose(1, 2)
        # (B, T*M, C)


        if padding_mask is not None:
            # Expand out mel and time dim
            features = features.reshape(B, T, M, C)
            # (B, T, M, C)
            # Downsample padding mask from frames to patches
            padding_mask = self.forward_padding_mask(features, padding_mask)
            # Expand (copy) time mask across associated freq patches
            padding_mask = padding_mask.unsqueeze(-1).expand(-1, -1, M)
            # Collapse mel and time dim
            padding_mask = padding_mask.reshape(B, -1)
            # (B, T*M)
            features = features.reshape(B, -1, C)
            # (B, T*M, C)

        if attn_mask is not None:
            padding_mask |= attn_mask

        # Apply layer wise normalisation
        features = self.layer_norm(features)

        # BUG: padding is computes over flattened spectrum
        # i.e. padding in time starts to mask some frequency tokens...
        # The above changes fixes this bug
        # if padding_mask is not None:
        #     padding_mask = self.forward_padding_mask(features, padding_mask)

        if self.post_extract_proj is not None:
            features = self.post_extract_proj(features)

        x = self.dropout_input(features)

        x, layer_results = self.encoder(
            x,
            padding_mask=padding_mask,
            layer=return_encoder_layer,
        )

        if self.predictor is not None:
            x = self.predictor_dropout(x)
            logits = self.predictor(x)

            if padding_mask is not None and padding_mask.any():
                logits[padding_mask] = 0
                logits = logits.sum(dim=1)
                logits = logits / (~padding_mask).sum(dim=1).unsqueeze(-1).expand_as(logits)
            else:
                logits = logits.mean(dim=1)

            lprobs = torch.sigmoid(logits)

            return lprobs, padding_mask
        else:
            if return_encoder_layer is not None:
                # layer_results -> [(x_i, attn)]
                # x_i : output of current layer fed to next layer
                # attn : attention weights, averaged over heads
                # index 0: is inputs to encoder layer
                return x, padding_mask, layer_results[1:]
            else:
                return x, padding_mask

    def extract_features(
            self,
            source: torch.Tensor,
            fbank_mean: float = 0,
            fbank_std: float = 0.5,
            padding_mask: Optional[torch.Tensor] = None,
            **kwargs,
    ):
        fbank = self.preprocess(source, fbank_mean=fbank_mean, fbank_std=fbank_std)

        if padding_mask is not None:
            padding_mask = self.forward_padding_mask(fbank, padding_mask)
        fbank = fbank.unsqueeze(1)
        return self.forward(fbank, padding_mask=padding_mask, **kwargs)




# --------------------------------------------------------
# BEATs: Audio Pre-Training with Acoustic Tokenizers (https://arxiv.org/abs/2212.09058)
# Github source: https://github.com/microsoft/unilm/tree/master/beats
# Copyright (c) 2022 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import math
import numpy as np
from typing import Dict, Optional, Tuple
import torch
from torch import Tensor, nn
import torch.nn.functional as F
from torch.nn import LayerNorm, Parameter
from .modules import (
    GradMultiply,
    SamePad,
    get_activation_fn,
    GLU_Linear,
    quant_noise,
)


class TransformerEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()

        self.dropout = args.dropout
        self.embedding_dim = args.encoder_embed_dim

        self.pos_conv = nn.Conv1d(
            self.embedding_dim,
            self.embedding_dim,
            kernel_size=args.conv_pos,
            padding=args.conv_pos // 2,
            groups=args.conv_pos_groups,
        )
        dropout = 0
        std = math.sqrt((4 * (1.0 - dropout)) / (args.conv_pos * self.embedding_dim))
        nn.init.normal_(self.pos_conv.weight, mean=0, std=std)
        nn.init.constant_(self.pos_conv.bias, 0)

        self.pos_conv = nn.utils.weight_norm(self.pos_conv, name="weight", dim=2)
        self.pos_conv = nn.Sequential(self.pos_conv, SamePad(args.conv_pos), nn.GELU())

        if hasattr(args, "relative_position_embedding"):
            self.relative_position_embedding = args.relative_position_embedding
            self.num_buckets = args.num_buckets
            self.max_distance = args.max_distance
        else:
            self.relative_position_embedding = False
            self.num_buckets = 0
            self.max_distance = 0

        self.layers = nn.ModuleList(
            [
                TransformerSentenceEncoderLayer(
                    embedding_dim=self.embedding_dim,
                    ffn_embedding_dim=args.encoder_ffn_embed_dim,
                    num_attention_heads=args.encoder_attention_heads,
                    dropout=self.dropout,
                    attention_dropout=args.attention_dropout,
                    activation_dropout=args.activation_dropout,
                    activation_fn=args.activation_fn,
                    layer_norm_first=args.layer_norm_first,
                    deep_norm=args.deep_norm,
                    has_relative_attention_bias=self.relative_position_embedding,
                    num_buckets=self.num_buckets,
                    max_distance=self.max_distance,
                    gru_rel_pos=args.gru_rel_pos,
                    encoder_layers=args.encoder_layers,
                )
                for i in range(args.encoder_layers)
            ]
        )
        if self.relative_position_embedding:
            for i in range(1, args.encoder_layers):
                del self.layers[i].self_attn.relative_attention_bias
                self.layers[i].self_attn.relative_attention_bias = self.layers[0].self_attn.relative_attention_bias

        self.layer_norm_first = args.layer_norm_first
        self.layer_norm = LayerNorm(self.embedding_dim)
        self.layerdrop = args.encoder_layerdrop

        self.apply(init_bert_params)

        if args.deep_norm:
            deep_norm_beta = math.pow(8 * args.encoder_layers, -1 / 4)
            for i in range(args.encoder_layers):
                nn.init.xavier_normal_(self.layers[i].self_attn.k_proj.weight, gain=1)
                nn.init.xavier_normal_(self.layers[i].self_attn.v_proj.weight, gain=deep_norm_beta)
                nn.init.xavier_normal_(self.layers[i].self_attn.q_proj.weight, gain=1)
                nn.init.xavier_normal_(self.layers[i].self_attn.out_proj.weight, gain=deep_norm_beta)
                nn.init.xavier_normal_(self.layers[i].fc1.weight, gain=deep_norm_beta)
                nn.init.xavier_normal_(self.layers[i].fc2.weight, gain=deep_norm_beta)

        self.layer_wise_gradient_decay_ratio = getattr(args, "layer_wise_gradient_decay_ratio", 1)

    def forward(self, x, padding_mask=None, layer=None):
        x, layer_results = self.extract_features(x, padding_mask, layer)

        if self.layer_norm_first and layer is None:
            x = self.layer_norm(x)

        return x, layer_results

    def extract_features(self, x, padding_mask=None, tgt_layer=None):

        if padding_mask is not None:
            x[padding_mask] = 0

        x_conv = self.pos_conv(x.transpose(1, 2))
        x_conv = x_conv.transpose(1, 2)
        x = x + x_conv

        if not self.layer_norm_first:
            x = self.layer_norm(x)

        x = F.dropout(x, p=self.dropout, training=self.training)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)

        layer_results = []
        z = None
        if tgt_layer is not None:
            layer_results.append((x, z))
        r = None
        pos_bias = None
        for i, layer in enumerate(self.layers):
            need_weight_current = not isinstance(tgt_layer, bool) and i == tgt_layer
            need_weight_all = isinstance(tgt_layer, bool) and tgt_layer
            if self.layer_wise_gradient_decay_ratio != 1.0:
                x = GradMultiply.apply(x, self.layer_wise_gradient_decay_ratio)
            dropout_probability = np.random.random()
            if not self.training or (dropout_probability > self.layerdrop):
                x, z, pos_bias = layer(x, self_attn_padding_mask=padding_mask, need_weights=need_weight_current or need_weight_all, pos_bias=pos_bias)
            if tgt_layer is not None:
                layer_results.append((x, z))
            if need_weight_current:
                r = x
                break

        if r is not None:
            x = r

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        return x, layer_results


class TransformerSentenceEncoderLayer(nn.Module):
    def __init__(
            self,
            embedding_dim: float = 768,
            ffn_embedding_dim: float = 3072,
            num_attention_heads: float = 8,
            dropout: float = 0.1,
            attention_dropout: float = 0.1,
            activation_dropout: float = 0.1,
            activation_fn: str = "relu",
            layer_norm_first: bool = False,
            deep_norm: bool = False,
            has_relative_attention_bias: bool = False,
            num_buckets: int = 0,
            max_distance: int = 0,
            rescale_init: bool = False,
            gru_rel_pos: bool = False,
            encoder_layers: int = 0,
    ) -> None:

        super().__init__()
        self.embedding_dim = embedding_dim
        self.dropout = dropout
        self.activation_dropout = activation_dropout

        self.activation_name = activation_fn
        self.activation_fn = get_activation_fn(activation_fn)
        self.self_attn = MultiheadAttention(
            self.embedding_dim,
            num_attention_heads,
            dropout=attention_dropout,
            self_attention=True,
            has_relative_attention_bias=has_relative_attention_bias,
            num_buckets=num_buckets,
            max_distance=max_distance,
            rescale_init=rescale_init,
            gru_rel_pos=gru_rel_pos,
        )

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(self.activation_dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.layer_norm_first = layer_norm_first

        self.self_attn_layer_norm = LayerNorm(self.embedding_dim)

        if self.activation_name == "glu":
            self.fc1 = GLU_Linear(self.embedding_dim, ffn_embedding_dim, "swish")
        else:
            self.fc1 = nn.Linear(self.embedding_dim, ffn_embedding_dim)
        self.fc2 = nn.Linear(ffn_embedding_dim, self.embedding_dim)

        self.final_layer_norm = LayerNorm(self.embedding_dim)

        self.deep_norm = deep_norm
        if self.deep_norm:
            self.deep_norm_alpha = math.pow(2 * encoder_layers, 1 / 4)
        else:
            self.deep_norm_alpha = 1

    def forward(
            self,
            x: torch.Tensor,
            self_attn_mask: torch.Tensor = None,
            self_attn_padding_mask: torch.Tensor = None,
            need_weights: bool = False,
            pos_bias=None
    ):
        residual = x

        if self.layer_norm_first:
            x = self.self_attn_layer_norm(x)
            x, attn, pos_bias = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                need_weights=False,
                attn_mask=self_attn_mask,
                position_bias=pos_bias
            )
            x = self.dropout1(x)
            x = residual + x

            residual = x
            x = self.final_layer_norm(x)
            if self.activation_name == "glu":
                x = self.fc1(x)
            else:
                x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual + x
        else:
            x, attn, pos_bias = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                need_weights=need_weights,
                attn_mask=self_attn_mask,
                position_bias=pos_bias
            )

            x = self.dropout1(x)
            x = residual * self.deep_norm_alpha + x

            x = self.self_attn_layer_norm(x)

            residual = x
            if self.activation_name == "glu":
                x = self.fc1(x)
            else:
                x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual * self.deep_norm_alpha + x
            x = self.final_layer_norm(x)

        return x, attn, pos_bias


class MultiheadAttention(nn.Module):
    """Multi-headed attention.

    See "Attention Is All You Need" for more details.
    """

    def __init__(
            self,
            embed_dim,
            num_heads,
            kdim=None,
            vdim=None,
            dropout=0.0,
            bias=True,
            add_bias_kv=False,
            add_zero_attn=False,
            self_attention=False,
            encoder_decoder_attention=False,
            q_noise=0.0,
            qn_block_size=8,
            has_relative_attention_bias=False,
            num_buckets=32,
            max_distance=128,
            gru_rel_pos=False,
            rescale_init=False,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.qkv_same_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout_module = nn.Dropout(dropout)

        self.has_relative_attention_bias = has_relative_attention_bias
        self.num_buckets = num_buckets
        self.max_distance = max_distance
        if self.has_relative_attention_bias:
            self.relative_attention_bias = nn.Embedding(num_buckets, num_heads)

        self.head_dim = embed_dim // num_heads
        self.q_head_dim = self.head_dim
        self.k_head_dim = self.head_dim
        assert (
                self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim ** -0.5

        self.self_attention = self_attention
        self.encoder_decoder_attention = encoder_decoder_attention

        assert not self.self_attention or self.qkv_same_dim, (
            "Self-attention requires query, key and " "value to be of the same size"
        )

        k_bias = True
        if rescale_init:
            k_bias = False

        k_embed_dim = embed_dim
        q_embed_dim = embed_dim

        self.k_proj = quant_noise(
            nn.Linear(self.kdim, k_embed_dim, bias=k_bias), q_noise, qn_block_size
        )
        self.v_proj = quant_noise(
            nn.Linear(self.vdim, embed_dim, bias=bias), q_noise, qn_block_size
        )
        self.q_proj = quant_noise(
            nn.Linear(embed_dim, q_embed_dim, bias=bias), q_noise, qn_block_size
        )

        self.out_proj = quant_noise(
            nn.Linear(embed_dim, embed_dim, bias=bias), q_noise, qn_block_size
        )

        if add_bias_kv:
            self.bias_k = Parameter(torch.Tensor(1, 1, embed_dim))
            self.bias_v = Parameter(torch.Tensor(1, 1, embed_dim))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self.gru_rel_pos = gru_rel_pos
        if self.gru_rel_pos:
            self.grep_linear = nn.Linear(self.q_head_dim, 8)
            self.grep_a = nn.Parameter(torch.ones(1, num_heads, 1, 1))

        self.reset_parameters()

    def reset_parameters(self):
        if self.qkv_same_dim:
            # Empirically observed the convergence to be much better with
            # the scaled initialization
            nn.init.xavier_uniform_(self.k_proj.weight, gain=1 / math.sqrt(2))
            nn.init.xavier_uniform_(self.v_proj.weight, gain=1 / math.sqrt(2))
            nn.init.xavier_uniform_(self.q_proj.weight, gain=1 / math.sqrt(2))
        else:
            nn.init.xavier_uniform_(self.k_proj.weight)
            nn.init.xavier_uniform_(self.v_proj.weight)
            nn.init.xavier_uniform_(self.q_proj.weight)

        nn.init.xavier_uniform_(self.out_proj.weight)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            nn.init.xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            nn.init.xavier_normal_(self.bias_v)
        if self.has_relative_attention_bias:
            nn.init.xavier_normal_(self.relative_attention_bias.weight)

    def _relative_positions_bucket(self, relative_positions, bidirectional=True):
        num_buckets = self.num_buckets
        max_distance = self.max_distance
        relative_buckets = 0

        if bidirectional:
            num_buckets = num_buckets // 2
            relative_buckets += (relative_positions > 0).to(torch.long) * num_buckets
            relative_positions = torch.abs(relative_positions)
        else:
            relative_positions = -torch.min(relative_positions, torch.zeros_like(relative_positions))

        max_exact = num_buckets // 2
        is_small = relative_positions < max_exact

        relative_postion_if_large = max_exact + (
                torch.log(relative_positions.float() / max_exact)
                / math.log(max_distance / max_exact)
                * (num_buckets - max_exact)
        ).to(torch.long)
        relative_postion_if_large = torch.min(
            relative_postion_if_large, torch.full_like(relative_postion_if_large, num_buckets - 1)
        )

        relative_buckets += torch.where(is_small, relative_positions, relative_postion_if_large)
        return relative_buckets

    def compute_bias(self, query_length, key_length):
        context_position = torch.arange(query_length, dtype=torch.long)[:, None]
        memory_position = torch.arange(key_length, dtype=torch.long)[None, :]
        relative_position = memory_position - context_position
        relative_position_bucket = self._relative_positions_bucket(
            relative_position,
            bidirectional=True
        )
        relative_position_bucket = relative_position_bucket.to(self.relative_attention_bias.weight.device)
        values = self.relative_attention_bias(relative_position_bucket)
        values = values.permute([2, 0, 1])
        return values

    def forward(
            self,
            query,
            key: Optional[Tensor],
            value: Optional[Tensor],
            key_padding_mask: Optional[Tensor] = None,
            incremental_state: Optional[Dict[str, Dict[str, Optional[Tensor]]]] = None,
            need_weights: bool = True,
            static_kv: bool = False,
            attn_mask: Optional[Tensor] = None,
            before_softmax: bool = False,
            need_head_weights: bool = False,
            position_bias: Optional[Tensor] = None
    ) -> Tuple[Tensor, Optional[Tensor], Optional[Tensor]]:
        """Input shape: Time x Batch x Channel

        Args:
            key_padding_mask (ByteTensor, optional): mask to exclude
                keys that are pads, of shape `(batch, src_len)`, where
                padding elements are indicated by 1s.
            need_weights (bool, optional): return the attention weights,
                averaged over heads (default: False).
            attn_mask (ByteTensor, optional): typically used to
                implement causal attention, where the mask prevents the
                attention from looking forward in time (default: None).
            before_softmax (bool, optional): return the raw attention
                weights and values before the attention softmax.
            need_head_weights (bool, optional): return the attention
                weights for each head. Implies *need_weights*. Default:
                return the average attention weights over all heads.
        """
        if need_head_weights:
            need_weights = True

        is_tpu = query.device.type == "xla"

        tgt_len, bsz, embed_dim = query.size()
        src_len = tgt_len
        assert embed_dim == self.embed_dim
        assert list(query.size()) == [tgt_len, bsz, embed_dim]
        if key is not None:
            src_len, key_bsz, _ = key.size()
            if not torch.jit.is_scripting():
                assert key_bsz == bsz
                assert value is not None
                assert src_len, bsz == value.shape[:2]

        if self.has_relative_attention_bias and position_bias is None:
            position_bias = self.compute_bias(tgt_len, src_len)
            position_bias = position_bias.unsqueeze(0).repeat(bsz, 1, 1, 1).view(bsz * self.num_heads, tgt_len, src_len)

        if incremental_state is not None:
            saved_state = self._get_input_buffer(incremental_state)
            if saved_state is not None and "prev_key" in saved_state:
                # previous time steps are cached - no need to recompute
                # key and value if they are static
                if static_kv:
                    assert self.encoder_decoder_attention and not self.self_attention
                    key = value = None
        else:
            saved_state = None

        if self.self_attention:
            q = self.q_proj(query)
            k = self.k_proj(query)
            v = self.v_proj(query)
        elif self.encoder_decoder_attention:
            # encoder-decoder attention
            q = self.q_proj(query)
            if key is None:
                assert value is None
                k = v = None
            else:
                k = self.k_proj(key)
                v = self.v_proj(key)

        else:
            assert key is not None and value is not None
            q = self.q_proj(query)
            k = self.k_proj(key)
            v = self.v_proj(value)
        q *= self.scaling
        alpha = 32
        q *= 1 / alpha

        if self.bias_k is not None:
            assert self.bias_v is not None
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        key_padding_mask.new_zeros(key_padding_mask.size(0), 1),
                    ],
                    dim=1,
                )

        q = (
            q.contiguous()
                .view(tgt_len, bsz * self.num_heads, self.q_head_dim)
                .transpose(0, 1)
        )
        if k is not None:
            k = (
                k.contiguous()
                    .view(-1, bsz * self.num_heads, self.k_head_dim)
                    .transpose(0, 1)
            )
        if v is not None:
            v = (
                v.contiguous()
                    .view(-1, bsz * self.num_heads, self.head_dim)
                    .transpose(0, 1)
            )

        if saved_state is not None:
            # saved states are stored with shape (bsz, num_heads, seq_len, head_dim)
            if "prev_key" in saved_state:
                _prev_key = saved_state["prev_key"]
                assert _prev_key is not None
                prev_key = _prev_key.view(bsz * self.num_heads, -1, self.head_dim)
                if static_kv:
                    k = prev_key
                else:
                    assert k is not None
                    k = torch.cat([prev_key, k], dim=1)
                src_len = k.size(1)
            if "prev_value" in saved_state:
                _prev_value = saved_state["prev_value"]
                assert _prev_value is not None
                prev_value = _prev_value.view(bsz * self.num_heads, -1, self.head_dim)
                if static_kv:
                    v = prev_value
                else:
                    assert v is not None
                    v = torch.cat([prev_value, v], dim=1)
            prev_key_padding_mask: Optional[Tensor] = None
            if "prev_key_padding_mask" in saved_state:
                prev_key_padding_mask = saved_state["prev_key_padding_mask"]
            assert k is not None and v is not None
            key_padding_mask = MultiheadAttention._append_prev_key_padding_mask(
                key_padding_mask=key_padding_mask,
                prev_key_padding_mask=prev_key_padding_mask,
                batch_size=bsz,
                src_len=k.size(1),
                static_kv=static_kv,
            )

            saved_state["prev_key"] = k.view(bsz, self.num_heads, -1, self.head_dim)
            saved_state["prev_value"] = v.view(bsz, self.num_heads, -1, self.head_dim)
            saved_state["prev_key_padding_mask"] = key_padding_mask
            # In this branch incremental_state is never None
            assert incremental_state is not None
            incremental_state = self._set_input_buffer(incremental_state, saved_state)
        assert k is not None
        assert k.size(1) == src_len

        # This is part of a workaround to get around fork/join parallelism
        # not supporting Optional types.
        if key_padding_mask is not None and key_padding_mask.dim() == 0:
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.add_zero_attn:
            assert v is not None
            src_len += 1
            k = torch.cat([k, k.new_zeros((k.size(0), 1) + k.size()[2:])], dim=1)
            v = torch.cat([v, v.new_zeros((v.size(0), 1) + v.size()[2:])], dim=1)
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        torch.zeros(key_padding_mask.size(0), 1).type_as(
                            key_padding_mask
                        ),
                    ],
                    dim=1,
                )

        attn_weights = torch.bmm(q, k.transpose(1, 2))
        attn_weights = (attn_weights - attn_weights.max(dim=-1, keepdim=True)[0]) * alpha
        attn_weights = self.apply_sparse_mask(attn_weights, tgt_len, src_len, bsz)

        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0)
            attn_weights += attn_mask

        if key_padding_mask is not None:
            # don't attend to padding symbols
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            if not is_tpu:
                attn_weights = attn_weights.masked_fill(
                    key_padding_mask.unsqueeze(1).unsqueeze(2).to(torch.bool),
                    float("-inf"),
                )
            else:
                attn_weights = attn_weights.transpose(0, 2)
                attn_weights = attn_weights.masked_fill(key_padding_mask, float("-inf"))
                attn_weights = attn_weights.transpose(0, 2)
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if before_softmax:
            return attn_weights, v, position_bias

        if position_bias is not None:
            attn_mask_rel_pos = position_bias
            if self.gru_rel_pos == 1:
                query_layer = q.view(bsz, self.num_heads, tgt_len, self.q_head_dim) * alpha / self.scaling
                _B, _H, _L, __ = query_layer.size()
                gate_a, gate_b = torch.sigmoid(self.grep_linear(query_layer).view(
                    _B, _H, _L, 2, 4).sum(-1, keepdim=False)).chunk(2, dim=-1)
                gate_a_1 = gate_a * (gate_b * self.grep_a - 1.0) + 2.0
                attn_mask_rel_pos = gate_a_1.view(bsz * self.num_heads, tgt_len, 1) * position_bias

            attn_mask_rel_pos = attn_mask_rel_pos.view(attn_weights.size())

            attn_weights = attn_weights + attn_mask_rel_pos

        attn_weights_float = F.softmax(
            attn_weights, dim=-1
        )
        attn_weights = attn_weights_float.type_as(attn_weights)
        attn_probs = self.dropout_module(attn_weights)

        assert v is not None
        attn = torch.bmm(attn_probs, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]
        attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
        attn = self.out_proj(attn)
        attn_weights: Optional[Tensor] = None
        if need_weights:
            attn_weights = attn_weights_float.view(
                bsz, self.num_heads, tgt_len, src_len
            ).transpose(1, 0)
            if not need_head_weights:
                # average attention weights over heads
                attn_weights = attn_weights.mean(dim=0)

        return attn, attn_weights, position_bias

    @staticmethod
    def _append_prev_key_padding_mask(
            key_padding_mask: Optional[Tensor],
            prev_key_padding_mask: Optional[Tensor],
            batch_size: int,
            src_len: int,
            static_kv: bool,
    ) -> Optional[Tensor]:
        # saved key padding masks have shape (bsz, seq_len)
        if prev_key_padding_mask is not None and static_kv:
            new_key_padding_mask = prev_key_padding_mask
        elif prev_key_padding_mask is not None and key_padding_mask is not None:
            new_key_padding_mask = torch.cat(
                [prev_key_padding_mask.float(), key_padding_mask.float()], dim=1
            )
        # During incremental decoding, as the padding token enters and
        # leaves the frame, there will be a time when prev or current
        # is None
        elif prev_key_padding_mask is not None:
            if src_len > prev_key_padding_mask.size(1):
                filler = torch.zeros(
                    (batch_size, src_len - prev_key_padding_mask.size(1)),
                    device=prev_key_padding_mask.device,
                )
                new_key_padding_mask = torch.cat(
                    [prev_key_padding_mask.float(), filler.float()], dim=1
                )
            else:
                new_key_padding_mask = prev_key_padding_mask.float()
        elif key_padding_mask is not None:
            if src_len > key_padding_mask.size(1):
                filler = torch.zeros(
                    (batch_size, src_len - key_padding_mask.size(1)),
                    device=key_padding_mask.device,
                )
                new_key_padding_mask = torch.cat(
                    [filler.float(), key_padding_mask.float()], dim=1
                )
            else:
                new_key_padding_mask = key_padding_mask.float()
        else:
            new_key_padding_mask = prev_key_padding_mask
        return new_key_padding_mask

    def _get_input_buffer(
            self, incremental_state: Optional[Dict[str, Dict[str, Optional[Tensor]]]]
    ) -> Dict[str, Optional[Tensor]]:
        result = self.get_incremental_state(incremental_state, "attn_state")
        if result is not None:
            return result
        else:
            empty_result: Dict[str, Optional[Tensor]] = {}
            return empty_result

    def _set_input_buffer(
            self,
            incremental_state: Dict[str, Dict[str, Optional[Tensor]]],
            buffer: Dict[str, Optional[Tensor]],
    ):
        return self.set_incremental_state(incremental_state, "attn_state", buffer)

    def apply_sparse_mask(self, attn_weights, tgt_len: int, src_len: int, bsz: int):
        return attn_weights


def init_bert_params(module):
    """
    Initialize the weights specific to the BERT Model.
    This overrides the default initializations depending on the specified arguments.
        1. If normal_init_linear_weights is set then weights of linear
           layer will be initialized using the normal distribution and
           bais will be set to the specified value.
        2. If normal_init_embed_weights is set then weights of embedding
           layer will be initialized using the normal distribution.
        3. If normal_init_proj_weights is set then weights of
           in_project_weight for MultiHeadAttention initialized using
           the normal distribution (to be validated).
    """

    def normal_(data):
        # with FSDP, module params will be on CUDA, so we cast them back to CPU
        # so that the RNG is consistent with and without FSDP
        data.copy_(
            data.cpu().normal_(mean=0.0, std=0.02).to(data.device)
        )

    if isinstance(module, nn.Linear):
        normal_(module.weight.data)
        if module.bias is not None:
            module.bias.data.zero_()
    if isinstance(module, nn.Embedding):
        normal_(module.weight.data)
        if module.padding_idx is not None:
            module.weight.data[module.padding_idx].zero_()
    if isinstance(module, MultiheadAttention):
        normal_(module.q_proj.weight.data)
        normal_(module.k_proj.weight.data)
        normal_(module.v_proj.weight.data)



from torch import nn, Tensor


def mean_with_attn(x: Tensor, mask: Tensor | None, dim: int) -> Tensor:
    """Compute mean along dimension, ignoring masked positions."""
    if mask is None:
        return x.mean(dim=dim)
    # Clone to avoid modifying input
    x = x.clone()
    # Expand mask to match x dimensions if needed
    while mask.ndim < x.ndim:
        mask = mask.unsqueeze(-1)
    mask = mask.expand_as(x)
    x[mask] = 0
    count = (~mask).sum(dim=dim, keepdim=True).float()
    return x.sum(dim=dim) / count.squeeze(dim)


class HardLabelPredictor(nn.Module):
    _MASK_TOKEN = 0
    def __init__(self, latent_proj, apply_mask=False):
        super().__init__()
        self.latent_proj = latent_proj
        self.apply_mask = apply_mask

    def forward(self, Z, attn_mask=None):
        B, T, M, E = Z.shape
        Z = Z.reshape(B, -1, E)
        # (B, T*M, E)
        if self.apply_mask and attn_mask is not None and attn_mask.any():
            # Require clone for backprop
            Z = Z.clone()
            Z[attn_mask] = self._MASK_TOKEN

        logits = self.latent_proj(Z) # E -> C
        # (B, T*M, C)
        logits = mean_with_attn(logits, attn_mask, dim=-2)
        # (B, C)
        return logits


"""
AERD - Automated Elephant Rumble Detection.

A deep learning model for detecting and classifying elephant vocalizations
using the BEATs audio foundation model.
"""

from .model import AERDClassifier
from .weights import AERD_Weights, aerd, aerd_ensemble
from .labels import LABEL_MAPS, get_class_labels
from .beats import (
    BEATs,
    BEATsConfig,
    build_beats_model,
    beats_preprocessing,
    beats_extract_features,
)

import importlib.metadata

try:
    __version__ = importlib.metadata.version("aerd")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"  # Fallback for development mode

__all__ = [
    # Main API
    "AERDClassifier",
    "AERD_Weights",
    "aerd",
    "aerd_ensemble",
    # Labels
    "LABEL_MAPS",
    "get_class_labels",
    # BEATs utilities
    "BEATs",
    "BEATsConfig",
    "build_beats_model",
    "beats_preprocessing",
    "beats_extract_features",
]



"""
AERD Classifier - Automated Elephant Rumble Detection.

A clean inference model wrapping BEATs backbone with classification head.
"""

from __future__ import annotations

from math import floor
from typing import Dict, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from .beats import BEATs, BEATsConfig, build_beats_model
from .utils import PatchwisePredictor


# Default BEATs config (from BEATs_iter3_plus_AS20K_finetuned_on_AS2M_cpt1)
_DEFAULT_BEATS_CONFIG = {
    "input_patch_size": 16,
    "embed_dim": 512,
    "conv_bias": False,
    "encoder_layers": 12,
    "encoder_embed_dim": 768,
    "encoder_ffn_embed_dim": 3072,
    "encoder_attention_heads": 12,
    "activation_fn": "gelu",
    "layer_wise_gradient_decay_ratio": 0.6,
    "layer_norm_first": False,
    "deep_norm": True,
    "dropout": 0.0,
    "attention_dropout": 0.0,
    "activation_dropout": 0.0,
    "encoder_layerdrop": 0.0,
    "dropout_input": 0.0,
    "conv_pos": 128,
    "conv_pos_groups": 16,
    "relative_position_embedding": True,
    "num_buckets": 320,
    "max_distance": 800,
    "gru_rel_pos": True,
    "finetuned_model": True,
    "predictor_dropout": 0.0,
    "predictor_class": 527,
}


class AERDClassifier(nn.Module):
    """
    AERD Classifier for elephant call detection and classification.

    Wraps a BEATs backbone with a classification head for inference.

    Args:
        num_classes: Number of output classes.
        fbank_mean: Mean for fbank normalization.
        fbank_std: Standard deviation for fbank normalization.
        predictor_type: Type of predictor head ('seq', 'label', 'seq-spec', 'seq-full').
        beats_config: BEATs configuration dict (uses default if None).
        class_labels: Mapping from class index to label name.
        dataset_info: Metadata about the training dataset and fold.
    """

    SAMPLE_RATE = 16_000

    def __init__(
        self,
        num_classes: int = 5,
        fbank_mean: float = 14.7274,
        fbank_std: float = 2.9366,
        predictor_type: str = "seq",
        beats_config: dict | None = None,
        class_labels: dict[int, str] | None = None,
        dataset_info: dict | None = None,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.fbank_mean = fbank_mean
        self.fbank_std = fbank_std
        self.predictor_type = predictor_type
        self.class_labels = class_labels
        self.dataset_info = dataset_info

        # Build BEATs backbone with config
        if beats_config is None:
            beats_config = _DEFAULT_BEATS_CONFIG.copy()

        cfg = BEATsConfig(beats_config)
        cfg.encoder_layerdrop = 0.0  # Disable for inference
        self.backbone = BEATs(cfg)
        self.backbone.predictor = None  # We use our own predictor

        # Compute number of spectral patches
        num_mel_bins = 128
        padding = self.backbone.patch_embedding.padding
        dilation = self.backbone.patch_embedding.dilation
        stride = self.backbone.patch_embedding.stride
        kernel_size = self.backbone.patch_embedding.kernel_size
        self.num_spectral_patches = int(
            floor(
                (num_mel_bins + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1)
                / stride[1]
                + 1
            )
        )

        # Build predictor head
        encoder_embed_dim = cfg.encoder_embed_dim
        self.predictor = self._build_predictor(predictor_type, encoder_embed_dim, num_classes)

    def _build_predictor(
        self, predictor_type: str, embed_dim: int, num_classes: int
    ) -> nn.Module:
        """Build the classification head based on predictor type."""
        if predictor_type == "seq":
            # Mean pool mel features then project
            return PatchwisePredictor(
                nn.Linear(embed_dim, num_classes),
                proj_over="mean_spec",
            )
        elif predictor_type == "label":
            # Single label per input
            from .utils import HardLabelPredictor
            return HardLabelPredictor(nn.Linear(embed_dim, num_classes))
        elif predictor_type == "seq-spec":
            # Flatten mel dim into embedding
            return PatchwisePredictor(
                nn.Linear(embed_dim * self.num_spectral_patches, num_classes),
                proj_over="all_spec",
            )
        elif predictor_type == "seq-full":
            # Prediction for each (T, M) patch
            return PatchwisePredictor(
                nn.Linear(embed_dim, num_classes),
                proj_over=None,
            )
        else:
            raise ValueError(f"Unknown predictor type: {predictor_type}")

    def preprocess(
        self,
        waveform: Tensor,
        padding_mask: Tensor | None = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Preprocess raw waveform to mel-spectrogram (fbank).

        Args:
            waveform: Raw audio tensor of shape (batch, samples) or (samples,).
            padding_mask: Optional boolean mask indicating padded positions.

        Returns:
            fbank: Mel-spectrogram of shape (batch, time, mel_bins).
            padding_mask: Padding mask for the fbank.
        """
        # Handle single waveform
        if waveform.ndim == 1:
            waveform = waveform.unsqueeze(0)

        batch_size, audio_len = waveform.shape

        # Compute fbank using BEATs preprocessing
        fbank = self.backbone.preprocess(
            waveform,
            fbank_mean=self.fbank_mean,
            fbank_std=self.fbank_std,
        )

        # Create or propagate padding mask
        if padding_mask is None:
            padding_mask = torch.zeros(
                batch_size, audio_len, device=waveform.device, dtype=torch.bool
            )

        padding_mask = self.backbone.forward_padding_mask(fbank, padding_mask)

        return fbank, padding_mask

    def forward(
        self,
        x: Tensor,
        padding_mask: Tensor | None = None,
        return_features: bool = False,
    ) -> Tensor | Tuple[Tensor, Tensor]:
        """
        Forward pass through the model.

        Args:
            x: Either raw waveform (batch, samples) or preprocessed fbank (batch, time, mel).
            padding_mask: Optional padding mask.
            return_features: If True, also return intermediate features.

        Returns:
            logits: Classification logits of shape (batch, time, num_classes) or (batch, num_classes).
            features: (optional) Intermediate features if return_features=True.
        """
        # Check if input is raw waveform or fbank
        if x.ndim == 2 and x.shape[-1] != 128:
            # Assume raw waveform (batch, samples)
            fbank, padding_mask = self.preprocess(x, padding_mask)
        elif x.ndim == 1:
            # Single raw waveform
            fbank, padding_mask = self.preprocess(x.unsqueeze(0), padding_mask)
        else:
            fbank = x

        # Add channel dimension for backbone
        if fbank.ndim == 3:
            fbank = fbank.unsqueeze(1)  # (batch, 1, time, mel)

        # Forward through backbone
        features, attn_mask = self.backbone.forward(fbank, padding_mask=padding_mask)

        batch_size = fbank.shape[0]

        # Reshape features: (batch, tokens, embed) -> (batch, time, spectral, embed)
        features = features.reshape(batch_size, -1, self.num_spectral_patches, 768)

        # Forward through predictor
        logits = self.predictor(features, attn_mask=attn_mask)

        if return_features:
            return logits, features
        return logits

    def predict(
        self,
        waveform: Tensor,
        output: str = "tensor",
    ) -> Tensor | Dict[str, Tensor]:
        """
        Convenience method for inference.

        Args:
            waveform: Raw audio tensor.
            output: Output format — "tensor" returns probability tensor,
                "dict" returns {label_name: probability_tensor}.

        Returns:
            Probabilities as tensor or label-keyed dict.
        """
        self.eval()
        with torch.no_grad():
            logits = self.forward(waveform)
            probs = torch.sigmoid(logits)

            if output == "tensor":
                return probs
            elif output == "dict":
                if self.class_labels is None:
                    raise ValueError(
                        "class_labels not available — cannot produce dict output. "
                        "Load model with AERD_Weights enum to include labels."
                    )
                return {
                    self.class_labels[i]: probs[..., i]
                    for i in range(self.num_classes)
                }
            else:
                raise ValueError(f"Unknown output format: {output!r}. Use 'tensor' or 'dict'.")

    def predict_labels(
        self,
        waveform: Tensor,
        threshold: float = 0.5,
    ) -> list[list[str]] | list[list[list[str]]]:
        """
        Return class labels with probability above threshold.

        Args:
            waveform: Raw audio tensor (batch, samples) or (samples,).
            threshold: Probability threshold (default 0.5).

        Returns:
            If predictor produces per-frame output (seq, seq-spec, seq-full):
                list[list[list[str]]] — labels per frame per batch item.
            If predictor produces single output (label):
                list[list[str]] — one label list per batch item.
        """
        if self.class_labels is None:
            raise ValueError(
                "class_labels not available — cannot produce label output."
            )
        probs = self.predict(waveform, output="tensor")

        above = probs >= threshold

        has_time = self.predictor_type in ("seq", "seq-spec", "seq-full")

        if has_time:
            # probs shape: (batch, time, classes)
            return [
                [
                    [self.class_labels[i] for i in frame.nonzero(as_tuple=True)[0].tolist()]
                    for frame in batch_item
                ]
                for batch_item in above
            ]
        else:
            # probs shape: (batch, classes)
            return [
                [self.class_labels[i] for i in row.nonzero(as_tuple=True)[0].tolist()]
                for row in above
            ]

    @classmethod
    def from_pretrained(
        cls,
        checkpoint_path: str,
        map_location: str | torch.device = "cpu",
    ) -> "AERDClassifier":
        """
        Load a pretrained AERD model from checkpoint.

        Args:
            checkpoint_path: Path to the checkpoint file.
            map_location: Device to load the model to.

        Returns:
            Loaded AERDClassifier model.
        """
        checkpoint = torch.load(checkpoint_path, map_location=map_location, weights_only=True)

        # Extract hyperparameters
        hparams = checkpoint.get("hyper_parameters", {})
        num_classes = hparams.get("num_classes", 5)
        fbank_mean = hparams.get("fbank_mean", 14.7274)
        fbank_std = hparams.get("fbank_std", 2.9366)
        predictor_type = hparams.get("predictor", "seq")

        # Extract BEATs config if available
        beats_config = checkpoint.get("beats_config", None)

        # Extract class labels and dataset info
        class_labels = checkpoint.get("class_labels", None)
        dataset_info = checkpoint.get("dataset_info", None)

        # Create model
        model = cls(
            num_classes=num_classes,
            fbank_mean=fbank_mean,
            fbank_std=fbank_std,
            predictor_type=predictor_type,
            beats_config=beats_config,
            class_labels=class_labels,
            dataset_info=dataset_info,
        )

        # Load state dict
        state_dict = checkpoint["state_dict"]
        model.load_state_dict(state_dict)

        return model



from functools import lru_cache
from typing import Dict, List, Tuple

import torch
from numpy import floor
from torch import Tensor
from torch.nn import Parameter
from torch.nn.utils.rnn import PackedSequence, pad_packed_sequence

try:
    from .unilm.beats import BEATs, BEATsConfig

except ImportError as e:
    print(f"BEATs submodule not found, please initialise git submodules: {e}")
    raise e


def build_beats_model(
    weights_path: str | None = None, disable_layerdrop: bool = True, **cfg
) -> BEATs:
    checkpoint = None
    if weights_path:
        assert not cfg
        checkpoint = torch.load(weights_path, weights_only=True)
        cfg = checkpoint["cfg"]

    if isinstance(cfg, dict):
        cfg = BEATsConfig(cfg)
    assert isinstance(cfg, BEATsConfig)
    # Useful when using DDP as layerdrop does not work in the distributed
    # case. Not sure why since the seeds are synced across all nodes but I
    # suspect it has to do with the gradient graph. Since the encoder
    # layers are dropped on a per sample basis not per batch and so some
    # outputs go through all encoders while others don't.
    # See issue: https://github.com/pytorch/pytorch/issues/42793
    # Setting `find_unused_parameters=True`, seems to also solve the issue
    # at the cost of longer training epochs
    if disable_layerdrop:
        cfg.encoder_layerdrop = 0.0  # set prob to 0, no chance of drop out

    model = BEATs(cfg)
    if weights_path:
        assert checkpoint
        model.load_state_dict(checkpoint["model"])

    # disable beats own predictor, only use as feature extractor
    model.predictor = None
    return model


@lru_cache(maxsize=1)
def get_num_spectral_patches(model, num_mel_bins=128):
    # Compute number of spectral patches
    padding = model.patch_embedding.padding
    dilation = model.patch_embedding.dilation
    stride = model.patch_embedding.stride
    kernel_size = model.patch_embedding.kernel_size
    W_out = floor(
        (num_mel_bins + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1)
        / stride[1]
        + 1
    )
    return int(W_out)


def beats_preprocessing(
    model: BEATs,
    X: torch.Tensor | PackedSequence,  # shape: (batch, audio_len, n_channels, *)
    padding_mask: None | torch.Tensor | List[int] = None,
) -> Tuple[Tensor, Tensor]:
    pad_len = 0
    if isinstance(X, PackedSequence):
        assert padding_mask is None
        X, seq_len = pad_packed_sequence(X, batch_first=True)
        audio_len = X.shape[1]
        # seq_len is the length of each unpadded seq
        # pad_len is the length of padding at the end of the seq
        pad_len = audio_len - seq_len
        # TODO: change seq_len instead, leave not that in AERD this is pad_len not seq_len
    # NOTE: check PackedSequence first as PackedSequence is a subtype of Tuple
    # If multiple args are given by X
    # assume, second arg is pad_len
    elif isinstance(X, tuple) or isinstance(X, list):
        assert padding_mask is None
        X, pad_len = X

    add_batch_dim = X.ndim == 1

    if X.ndim > 2:
        batch_size, audio_len, *_ = X.shape
        # Flatten all dims, except batch
        X = X.flatten(start_dim=1)

    elif add_batch_dim:
        X = X.unsqueeze(0)

    X = X.contiguous()

    batch_size, audio_len = X.shape

    # If pad_len not set,
    # assume no padding
    # i.e. not attention masking
    if isinstance(pad_len, int):
        pad_len = torch.full((batch_size,), fill_value=pad_len, dtype=torch.long)

    fbank = model.preprocess(X)
    if padding_mask is None:
        # Ensure padding mask on same device as x
        # see https://lightning.ai/docs/pytorch/stable/accelerators/accelerator_prepare.html#init-tensors-using-tensor-to-and-register-buffer
        padding_mask = torch.zeros(
            batch_size, audio_len, device=X.device, dtype=torch.bool
        )
        # Mark all samples that correspond to padding tokens
        for i, pl in enumerate(pad_len):
            if pl == 0:
                continue
            padding_mask[i, -pl:] = True

        padding_mask = model.forward_padding_mask(fbank, padding_mask)

    # Add back empty channel dim, expected by backbone/patch embedding
    # fbank = fbank.unsqueeze(1)
    # Remove channel dim
    # padding_mask = padding_mask.squeeze()
    assert isinstance(padding_mask, Tensor)

    if add_batch_dim:
        fbank = fbank.squeeze(0)
        padding_mask = padding_mask.squeeze(0)

    return fbank, padding_mask


def beats_extract_features(
    model: BEATs,
    X: Tensor | Tuple[Tensor, Tensor],
    # return_spectral_patch=False,  # Return spectral patch dim, or take mean
    # return_fbank=False,
    padding_mask: Tensor | None = None,
    *,
    apply_mask: bool = True,
    return_encoder_layer: int | None = None,
    random_feature_mask: bool = False,
    mask_token: float | Tensor | Parameter = 0.0,
) -> Tuple[Tensor, Dict]:
    # Optional return dict
    opt_ret = dict()
    # X contains multiple arguments
    X_is_multi_arg = isinstance(X, tuple) or isinstance(X, list)

    if padding_mask is not None and X_is_multi_arg:
        fbank, _ = X
    elif X_is_multi_arg:
        fbank, padding_mask = X
    else:
        fbank = X
    assert isinstance(fbank, Tensor)

    add_batch_dim = bool(fbank.ndim == 2)
    if add_batch_dim:
        # Add batch dim
        fbank = fbank.unsqueeze(0)
        # Add channel dim
        fbank = fbank.unsqueeze(1)
    elif fbank.ndim == 3:
        # Add channel dim
        fbank = fbank.unsqueeze(1)

    if padding_mask is None:
        padding_mask = torch.zeros(fbank.size(-3), fbank.size(-2))
    assert isinstance(padding_mask, Tensor)

    if add_batch_dim:
        assert padding_mask.ndim == 1
        padding_mask = padding_mask.unsqueeze(0)

    assert (
        fbank.ndim == 4
    ), "fbanks must be of shape: (batch, nchannels, time_dim, mel_dim)"
    assert padding_mask is None or padding_mask.size(0) == fbank.size(
        0
    ), "batch_size must be equal for padding_mask and fbanks"
    assert not padding_mask.all(dim=1).any(), "batch contains sample with just padding"

    batch_size, nchannels, time_dim, mel_dim = fbank.shape

    if random_feature_mask:
        raise NotImplementedError
        feat_mask = feature_mask(
            fbank.size(),
            threshold=self.feature_mask_threshold,
            kernel_size=self.feature_mask_kernel_size,
            device=fbank.device,
        )
        fbank = fbank * feat_mask

    if return_encoder_layer is not None:
        # layer_results -> [(x_i, attn)]
        # x_i : output of current layer fed to next layer
        # attn : attention weights, averaged over heads
        # index 0: is inputs to encoder layer
        Z, attn_mask, encoder_layers = model.forward(
            fbank,
            padding_mask=padding_mask,
            return_encoder_layer=return_encoder_layer,
        )
        if return_encoder_layer is True:
            opt_ret["encoder_layers"] = encoder_layers
        elif isinstance(return_encoder_layer, int):
            opt_ret["encoder_layer"] = return_encoder_layer
            Z, _mask = encoder_layers[return_encoder_layer]
            Z = Z.transpose(0, 1)
            attn_mask = None
        else:
            raise ValueError("Unreachable")
    else:
        Z, attn_mask = model.forward(
            fbank,
            padding_mask=padding_mask,
        )

    assert isinstance(Z, Tensor)
    assert isinstance(attn_mask, Tensor) or attn_mask is None

    apply_mask = bool(apply_mask and (attn_mask is not None) and attn_mask.any())
    if apply_mask:
        # Mask output tokens
        # Note: these tokens contain random info, as they were masked to the encoder
        Z = Z.clone()  # Expensive copy needed for backprop graph
        # Can't modify view inplace: https://discuss.pytorch.org/t/leaf-variable-was-used-in-an-inplace-operation/308
        # Z = Z * attn_mask
        Z[attn_mask] = mask_token
        # Masked tensor impl: shapes through linear layer end up breaking
        # Z = masked_tensor(Z, attn_mask.unsqueeze(-1).expand_as(Z))

    num_spectral_patches = get_num_spectral_patches(model)
    # TODO: get from config
    num_features = 768
    # Reshape the hidden embedding tokens Z
    # (batch_size, num_tokens_temporal, num_tokens_spectral, num_features)
    # TODO: hard codeded feature dim
    Z = Z.reshape(batch_size, -1, num_spectral_patches, num_features)

    # Remove batch dim if one was inserted
    if add_batch_dim:
        Z = Z.squeeze(0)
        attn_mask = attn_mask.squeeze(0) if attn_mask is not None else None

    # If mask was applied, store attn_mask
    if apply_mask:
        opt_ret["attn_mask"] = attn_mask

    return Z, opt_ret



"""
Known label mappings for AERD models.

Label schemes follow the Poole 2011 taxonomy. The ann_ids are assigned
sequentially based on which labels appear in each dataset.
"""

from __future__ import annotations

from typing import Dict, Optional, Tuple

# Known label maps keyed by (dataset, variant)
LABEL_MAPS: Dict[Tuple[str, str], Dict[int, str]] = {
    # EV (vocalisation_type='call') — 5 classes, 5 folds
    ("elephant-voices", "call"): {
        0: "not-call",
        1: "rumble",
        2: "roar",
        3: "cry",
        4: "trumpet",
    },
    # EV (vocalisation_type='subcall') — 8 classes, 5 folds
    ("elephant-voices", "subcall"): {
        0: "not-call",
        1: "comment-rumble",
        2: "conflict-roar",
        3: "estrous-rumble",
        4: "female-chorus",
        5: "grumbling-rumble",
        6: "pulsated-play-trumpet",
        7: "trumpet-blast",
    },
    # LDC (annotation_scheme='poole') — 6 classes, 10 folds
    ("asian-voc", "poole"): {
        0: "not-call",
        1: "rumble",
        2: "roar",
        3: "bark",
        4: "trumpet",
        5: "squelch",
    },
    # Binary variants (any dataset)
    ("elephant-voices", "binary"): {0: "not-call", 1: "call"},
    ("asian-voc", "binary"): {0: "not-call", 1: "call"},
}

BINARY_LABELS: Dict[int, str] = {0: "not-call", 1: "call"}


def fix_target_mapping(
    target_mapping: Dict[str, int],
    is_binary: bool = False,
) -> Dict[int, str]:
    """Invert and fix a target_mapping from the training code.

    The training code's reverse mapping produces ``"call"`` at ID 1 (dict
    ordering overwrites ``"rumble"``).  For non-binary models the correct
    semantic label is ``"rumble"``, so this function corrects that.

    Args:
        target_mapping: Forward mapping ``{label_name: ann_id}`` from the
            training datamodule's ``target_mapping`` attribute.
        is_binary: When ``True`` the mapping is for a binary model and
            ``"call"`` is the correct label — no renaming is applied.

    Returns:
        Inverted mapping ``{ann_id: label_name}`` with the fix applied.
    """
    # Invert: {name: id} -> {id: name}
    inverted: Dict[int, str] = {v: k for k, v in target_mapping.items()}

    # Fix the "call" -> "rumble" collision for non-binary models
    if not is_binary and inverted.get(1) == "call":
        inverted[1] = "rumble"

    return inverted


def get_class_labels(
    dataset: str,
    variant: str,
) -> Optional[Dict[int, str]]:
    """Look up hardcoded class labels for a known (dataset, variant) pair.

    Args:
        dataset: Dataset name (``"elephant-voices"`` or ``"asian-voc"``).
        variant: Variant name (``"call"``, ``"subcall"``, ``"poole"``,
            or ``"binary"``).

    Returns:
        Label mapping ``{ann_id: label_name}`` or ``None`` if the
        combination is unknown.
    """
    return LABEL_MAPS.get((dataset, variant))



"""
AERD Model Weights Management.

Provides utilities for loading pretrained AERD model weights from GitHub Releases.
Supports per-fold and ensemble loading for k-fold cross-validation models.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Callable, List, Tuple

import torch
from torch.hub import load_state_dict_from_url

from .labels import LABEL_MAPS

if TYPE_CHECKING:
    from .model import AERDClassifier


# GitHub Release URL template
# Update this URL when publishing new releases
_WEIGHTS_URL_BASE = "https://github.com/CMGeldenhuys/AERD_inference/releases/download"


class AERD_Weights(Enum):
    """
    Available pretrained weights for AERD models.

    Each weight variant contains:
    - url_template: URL template with {version} and {fold} placeholders
    - version: Default version tag
    - num_classes: Number of output classes
    - num_folds: Number of k-fold splits
    - class_labels: Mapping from class index to label name
    - meta: Additional metadata (dataset, variant, etc.)
    """

    EV_CALL = {
        "url_template": f"{_WEIGHTS_URL_BASE}/v{{version}}/aerd_ev_call_ko{{fold}}_v{{version}}-{{hash}}.pt",
        "version": "1",
        "num_classes": 5,
        "num_folds": 5,
        "class_labels": LABEL_MAPS[("elephant-voices", "call")],
        "file_hashes": {
            1: "5876c7c3", 2: "75efa98e", 3: "5cc3c744",
            4: "48b2294d", 5: "9420e0de",
        },
        "meta": {
            "dataset": "elephant-voices",
            "variant": "call",
            "description": "AERD model for elephant call-type classification (EV dataset)",
        },
    }

    EV_SUBCALL = {
        "url_template": f"{_WEIGHTS_URL_BASE}/v{{version}}/aerd_ev_subcall_ko{{fold}}_v{{version}}-{{hash}}.pt",
        "version": "1",
        "num_classes": 8,
        "num_folds": 5,
        "class_labels": LABEL_MAPS[("elephant-voices", "subcall")],
        "file_hashes": {
            1: "060becac", 2: "20877eaa", 3: "8e3b703e",
            4: "38886767", 5: "b640338b",
        },
        "meta": {
            "dataset": "elephant-voices",
            "variant": "subcall",
            "description": "AERD model for elephant sub-call classification (EV dataset)",
        },
    }

    EV_BINARY = {
        "url_template": f"{_WEIGHTS_URL_BASE}/v{{version}}/aerd_ev_binary_ko{{fold}}_v{{version}}-{{hash}}.pt",
        "version": "1",
        "num_classes": 2,
        "num_folds": 5,
        "class_labels": LABEL_MAPS[("elephant-voices", "binary")],
        "file_hashes": {
            1: "d583165d", 2: "0c7fe335", 3: "4ca10435",
            4: "97ae4718", 5: "eceda608",
        },
        "meta": {
            "dataset": "elephant-voices",
            "variant": "binary",
            "description": "AERD binary call detection model (EV dataset)",
        },
    }

    LDC_POOLE = {
        "url_template": f"{_WEIGHTS_URL_BASE}/v{{version}}/aerd_ldc_poole_ko{{fold}}_v{{version}}-{{hash}}.pt",
        "version": "1",
        "num_classes": 6,
        "num_folds": 10,
        "class_labels": LABEL_MAPS[("asian-voc", "poole")],
        "file_hashes": {
            1: "481a3595", 2: "768a7871", 3: "fe8caa30",
            4: "3125e47e", 5: "f63a2396", 6: "d1dbcf1e",
            7: "ee90ff8a", 8: "f4ed2466", 9: "90729160",
            10: "94261f9c",
        },
        "meta": {
            "dataset": "asian-voc",
            "variant": "poole",
            "description": "AERD model for Asian elephant call classification (LDC dataset, Poole taxonomy)",
        },
    }

    LDC_BINARY = {
        "url_template": f"{_WEIGHTS_URL_BASE}/v{{version}}/aerd_ldc_binary_ko{{fold}}_v{{version}}-{{hash}}.pt",
        "version": "1",
        "num_classes": 2,
        "num_folds": 10,
        "class_labels": LABEL_MAPS[("asian-voc", "binary")],
        "file_hashes": {
            1: "393ba94d", 2: "72245d4f", 3: "309f3e8c",
            4: "3cb18507", 5: "ee816aac", 6: "61ea48e5",
            7: "dfd66fd1", 8: "7a59b271", 9: "69e1c0ca",
            10: "b9f3883d",
        },
        "meta": {
            "dataset": "asian-voc",
            "variant": "binary",
            "description": "AERD binary call detection model (LDC dataset)",
        },
    }

    DEFAULT_EV = EV_CALL      # Alias → loads EV call classification (fold 1 by default)
    DEFAULT_LDC = LDC_POOLE   # Alias → loads LDC Poole classification (fold 1 by default)
    ENSEMBLE_EV = EV_CALL     # Alias → semantic hint for ensemble use with aerd_ensemble()
    ENSEMBLE_LDC = LDC_POOLE  # Alias → semantic hint for ensemble use with aerd_ensemble()

    @property
    def url_template(self) -> str:
        return self.value["url_template"]

    @property
    def version(self) -> str:
        return self.value["version"]

    @property
    def num_classes(self) -> int:
        return self.value["num_classes"]

    @property
    def num_folds(self) -> int:
        return self.value["num_folds"]

    @property
    def class_labels(self) -> dict[int, str]:
        return self.value["class_labels"]

    @property
    def meta(self) -> dict:
        return self.value["meta"]

    @property
    def file_hashes(self) -> dict[int, str]:
        return self.value.get("file_hashes", {})

    def get_url(self, fold: int, version: str | None = None) -> str:
        """Build the download URL for a specific fold and version."""
        v = version or self.version
        file_hash = self.file_hashes.get(fold, "")
        return self.url_template.format(version=v, fold=fold, hash=file_hash)


def _load_weights(
    weights: AERD_Weights | str | None,
    fold: int = 1,
    progress: bool = True,
    check_hash: bool = False,
) -> dict | None:
    """
    Load weights from URL or local path.

    Args:
        weights: Weights specification (AERD_Weights enum, URL string, or local path).
        fold: Fold number (only used with AERD_Weights enum).
        progress: Show download progress bar.
        check_hash: Verify file hash after download.

    Returns:
        Loaded checkpoint dictionary, or None if weights is None.
    """
    if weights is None:
        return None

    if isinstance(weights, AERD_Weights):
        url = weights.get_url(fold)
        return load_state_dict_from_url(
            url,
            progress=progress,
            check_hash=check_hash,
            map_location="cpu",
            weights_only=True,
        )
    elif isinstance(weights, str):
        if weights.startswith(("http://", "https://")):
            return load_state_dict_from_url(
                weights,
                progress=progress,
                check_hash=check_hash,
                map_location="cpu",
                weights_only=True,
            )
        else:
            # Local file path
            return torch.load(weights, map_location="cpu", weights_only=True)
    else:
        raise TypeError(f"weights must be AERD_Weights, str, or None, got {type(weights)}")


def _build_model_from_checkpoint(
    checkpoint: dict,
    weights: AERD_Weights | str | None = None,
) -> "AERDClassifier":
    """Build an AERDClassifier from a loaded checkpoint dict."""
    from .model import AERDClassifier

    hparams = checkpoint.get("hyper_parameters", {})
    beats_config = checkpoint.get("beats_config", None)
    class_labels = checkpoint.get("class_labels", None)
    dataset_info = checkpoint.get("dataset_info", None)

    # Get metadata from weights enum if available
    if isinstance(weights, AERD_Weights):
        meta = weights.meta
        num_classes = weights.num_classes
        if class_labels is None:
            class_labels = weights.class_labels
    else:
        meta = {}
        num_classes = hparams.get("num_classes", 5)

    model = AERDClassifier(
        num_classes=num_classes,
        fbank_mean=hparams.get("fbank_mean", meta.get("fbank_mean", 14.7274)),
        fbank_std=hparams.get("fbank_std", meta.get("fbank_std", 2.9366)),
        predictor_type=hparams.get("predictor", meta.get("predictor", "seq")),
        beats_config=beats_config,
        class_labels=class_labels,
        dataset_info=dataset_info,
    )

    model.load_state_dict(checkpoint["state_dict"])
    model.eval()
    return model


def aerd(
    weights: AERD_Weights | str = AERD_Weights.DEFAULT_EV,
    fold: int = 1,
    progress: bool = True,
) -> Tuple["AERDClassifier", Callable]:
    """
    Load AERD model with pretrained weights.

    This is the main entry point for PyTorch Hub.

    Args:
        weights: Pretrained weights to load. Can be:
            - AERD_Weights.DEFAULT_EV: Default pretrained weights (EV call classification)
            - A URL string to download weights from
            - A local file path to load weights from
        fold: Fold number to load (1-indexed). Only used with AERD_Weights enum.
        progress: Show download progress bar.

    Returns:
        model: Loaded AERDClassifier model in eval mode.
        preprocess: Preprocessing function for audio input.

    Example:
        >>> import torch
        >>> model, preprocess = torch.hub.load("CMGeldenhuys/AERD_inference", "aerd")
        >>> audio = torch.randn(1, 16000 * 10)  # 10 seconds at 16kHz
        >>> logits = model(audio)
    """
    from .model import AERDClassifier

    checkpoint = _load_weights(weights, fold=fold, progress=progress)

    if checkpoint is None:
        model = AERDClassifier()
        model.eval()
    else:
        model = _build_model_from_checkpoint(checkpoint, weights)

    return model, model.preprocess


def aerd_ensemble(
    weights: AERD_Weights = AERD_Weights.DEFAULT_EV,
    folds: list[int] | None = None,
    progress: bool = True,
) -> Tuple[List["AERDClassifier"], Callable]:
    """
    Load an ensemble of AERD models across k-fold splits.

    Args:
        weights: Weight variant to load.
        folds: List of fold numbers to load (1-indexed). If None, loads all folds.
        progress: Show download progress bar.

    Returns:
        models: List of loaded AERDClassifier models in eval mode.
        preprocess: Preprocessing function (shared across all models).

    Example:
        >>> models, preprocess = aerd_ensemble(AERD_Weights.EV_CALL)
        >>> audio = torch.randn(1, 16000 * 10)
        >>> # Average predictions across folds
        >>> logits = torch.stack([m(audio) for m in models]).mean(dim=0)
    """
    if folds is None:
        folds = list(range(1, weights.num_folds + 1))

    models = []
    for fold in folds:
        checkpoint = _load_weights(weights, fold=fold, progress=progress)
        if checkpoint is not None:
            model = _build_model_from_checkpoint(checkpoint, weights)
            models.append(model)

    if not models:
        raise RuntimeError(f"No models loaded for {weights.name}")

    return models, models[0].preprocess



from .beats_helper import beats_extract_features, beats_preprocessing, build_beats_model
from .unilm.beats import BEATs, BEATsConfig

__all__ = [
    "BEATs",
    "BEATsConfig",
    "build_beats_model",
    "beats_extract_features",
    "beats_preprocessing",
]
